==> ppcboot-0.9.2.tar.bz2: ppcboot-0.9.2/board/cu824/flash.c <==
/*
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc8240.h>
#include <asm/processor.h>

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

#define FLASH_BANK_SIZE 0x800000
#define MAIN_SECT_SIZE  0x40000
#define PARAM_SECT_SIZE 0x8000

flash_info_t    flash_info[CFG_MAX_FLASH_BANKS];

static int write_data (flash_info_t *info, ulong dest, uchar *data);
static void write_via_fpu(volatile ulong *addr, uchar *data);
static __inline__ unsigned long get_msr(void);
static __inline__ void set_msr(unsigned long msr);


/*-----------------------------------------------------------------------
 */

unsigned long flash_init(void)
{
	int i, j;
	unsigned long size = 0;

	for (i = 0; i < CFG_MAX_FLASH_BANKS; i++)
	{
		flash_info[i].flash_id = 0x008988F4;
		flash_info[i].size = FLASH_BANK_SIZE;
		flash_info[i].sector_count = CFG_MAX_FLASH_SECT;
		memset(flash_info[i].protect, 0, CFG_MAX_FLASH_SECT);
		for (j = 0; j < flash_info[i].sector_count; j++)
		{
			if (j <= 7)
			{
				flash_info[i].start[j] = CFG_FLASH_BASE + i * FLASH_BANK_SIZE +
				                         j * PARAM_SECT_SIZE;
			}
			else
			{
				flash_info[i].start[j] = CFG_FLASH_BASE + i * FLASH_BANK_SIZE +
				                         (j - 7) * MAIN_SECT_SIZE;
			}
		}
		size += flash_info[i].size;
	}

	/* Protect monitor and environment sectors
	 */
#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
	flash_protect(FLAG_PROTECT_SET,
	              CFG_MONITOR_BASE,
	              CFG_MONITOR_BASE + CFG_MONITOR_LEN - 1,
	              &flash_info[0]);
#endif

#if (CFG_ENV_IS_IN_FLASH == 1) && defined(CFG_ENV_ADDR)
	flash_protect(FLAG_PROTECT_SET,
	              CFG_ENV_ADDR,
	              CFG_ENV_ADDR + CFG_ENV_SIZE - 1,
	              &flash_info[0]);
#endif

	return size;
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
	int i;

	switch (info->flash_id & FLASH_VENDMASK)
	{
		case (INTEL_MANUFACT & FLASH_VENDMASK):
			printf("Intel: ");
		break;
		default:
			printf("Unknown Vendor ");
			break;
	}

	switch (info->flash_id & FLASH_TYPEMASK)
	{
		case 0x88F4:
			printf("28F160F3B (16Mbit)\n");
			break;
		default:
			printf ("Unknown Chip Type\n");
			break;
	}

	printf("  Size: %ld MB in %d Sectors\n",
	       info->size >> 20, info->sector_count);

	printf("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; i++)
	{
		if ((i % 5) == 0)
		{
			printf ("\n   ");
		}
		printf (" %08lX%s", info->start[i],
		        info->protect[i] ? " (RO)" : "     ");
	}
	printf ("\n");
}

/*-----------------------------------------------------------------------
 */

void	flash_erase (flash_info_t *info, int s_first, int s_last)
{
	int flag, prot, sect;
	ulong start, now, last;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return;
	}

	if ((info->flash_id & FLASH_VENDMASK) !=
	    (INTEL_MANUFACT & FLASH_VENDMASK)) {
		printf ("Can erase only Intel flash types - aborted\n");
		return;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	start = get_timer (0);
	last  = start;
	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			vu_long *addr = (vu_long *)(info->start[sect]);
			unsigned long status;

			/* Disable interrupts which might cause a timeout here */
			flag = disable_interrupts();

			*addr = 0x00500050;	/* clear status register */
			*addr = 0x00200020;	/* erase setup */
			*addr = 0x00D000D0;	/* erase confirm */

			/* re-enable interrupts if necessary */
			if (flag)
				enable_interrupts();

			/* wait at least 80us - let's wait 1 ms */
			udelay (1000);

			while (((status = *addr) & 0x00800080) != 0x00800080) {
				if ((now=get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
					*addr = 0x00B000B0; /* suspend erase	  */
					*addr = 0x00FF00FF; /* reset to read mode */
					return;
				}

				/* show that we're waiting */
				if ((now - last) > 1000) {	/* every second */
					putc ('.');
					last = now;
				}
			}

			*addr = 0x00FF00FF;	/* reset to read mode */
		}
	}
	printf (" done\n");
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 4 - Flash not identified
 */

#define	FLASH_WIDTH	8	/* flash bus width in bytes */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong wp, msr;
	int l, rc;
	unsigned char data[8];
	u_long *datah = (u_long *)&data[0];
	u_long *datal = (u_long *)&data[4];

	if (info->flash_id == FLASH_UNKNOWN) {
		return 4;
	}

	wp = (addr & ~(FLASH_WIDTH-1));	/* get lower FLASH_WIDTH aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
#if 0
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<FLASH_WIDTH && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<FLASH_WIDTH; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}

		if ((rc = write_data(info, wp, data)) != 0) {
			return (rc);
		}
		wp += FLASH_WIDTH;
#else
		return ERR_ALIGN;
#endif
	}

#if 1
	if (cnt % FLASH_WIDTH != 0)
	{
		return ERR_ALIGN;
	}
#endif

	/*
	 * handle FLASH_WIDTH aligned part
	 */
	msr = get_msr();
	set_msr(msr | MSR_FP);
	while (cnt >= FLASH_WIDTH) {
		*datah = *(u_long *)src;
		*datal = *(u_long *)(src + 4);
		if ((rc = write_data(info, wp, data)) != 0) {
			set_msr(msr);
			return (rc);
		}
		wp  += FLASH_WIDTH;
		cnt -= FLASH_WIDTH;
		src += FLASH_WIDTH;
	}
	set_msr(msr);

	if (cnt == 0) {
		return (0);
	}

#if 0
	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<FLASH_WIDTH && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i<FLASH_WIDTH; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}

	return (write_data(info, wp, data));
#else
	return 0;
#endif
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_data (flash_info_t *info, ulong dest, uchar *data)
{
	vu_long *addr = (vu_long *)dest;
	ulong status;
	ulong start;
	ulong *hdata = (ulong *)&data[0];
	ulong *ldata = (ulong *)&data[4];
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*addr & *hdata) != *hdata ||
	    (*(addr + 1) & *ldata) != *ldata) {
		return (2);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	*addr = 0x00400040;		/* write setup */
	write_via_fpu(addr, data);

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	start = get_timer (0);

	while ((((status = *addr) & 0x00800080) != 0x00800080) ||
	      ((*(addr + 1)) & 0x00800080) != 0x00800080) {
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			*addr = 0x00FF00FF;	/* restore read mode */
			return (1);
		}
	}

	*addr = 0x00FF00FF;	/* restore read mode */

	return (0);
}

/*-----------------------------------------------------------------------
 */
static void write_via_fpu(volatile ulong *addr, uchar *data)
{
__asm__ __volatile__ (
	"lfd  1, 0(%0)\n
	 stfd 1, 0(%1)\n"
	 :
	 : "r" (data), "r" (addr)
);
}
/*-----------------------------------------------------------------------
 */
static __inline__ unsigned long get_msr(void)
{
    unsigned long msr;

    asm volatile("mfmsr %0" : "=r" (msr) :);
    return msr;
}

static __inline__ void set_msr(unsigned long msr)
{
    asm volatile("mtmsr %0" : : "r" (msr));
}

==> ppcboot-1.0.1.tar.bz2: ppcboot-1.0.1/board/cu824/flash.c <==
/*
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc8240.h>
#include <asm/processor.h>

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

#define FLASH_BANK_SIZE 0x800000
#define MAIN_SECT_SIZE  0x40000
#define PARAM_SECT_SIZE 0x8000

flash_info_t    flash_info[CFG_MAX_FLASH_BANKS];

static int write_data (flash_info_t *info, ulong dest, uchar *data);
static void write_via_fpu(volatile ulong *addr, uchar *data);
static __inline__ unsigned long get_msr(void);
static __inline__ void set_msr(unsigned long msr);


/*-----------------------------------------------------------------------
 */

unsigned long flash_init(void)
{
	int i, j;
	unsigned long size = 0;

	for (i = 0; i < CFG_MAX_FLASH_BANKS; i++)
	{
		*(volatile u_long *)(CFG_FLASH_BASE + i * FLASH_BANK_SIZE) =
		                                      0x00900090;
		if (*(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE) ==
		    *(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE + 4) &&
		    *(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE) ==
		    0x00890089 &&
		    *(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE + 8) ==
		    *(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE + 12) &&
		    *(volatile u_long *)(CFG_FLASH_BASE +
		                         i * FLASH_BANK_SIZE + 8) ==
		    0x88f488f4)
		{
			flash_info[i].flash_id = 0x008988F4;
		}
		else
		{
			flash_info[i].flash_id = FLASH_UNKNOWN;
			goto Done;
		}
		*(volatile u_long *)(CFG_FLASH_BASE + i * FLASH_BANK_SIZE) =
		                                      0xffffffff;
		flash_info[i].size = FLASH_BANK_SIZE;
		flash_info[i].sector_count = CFG_MAX_FLASH_SECT;
		memset(flash_info[i].protect, 0, CFG_MAX_FLASH_SECT);
		for (j = 0; j < flash_info[i].sector_count; j++)
		{
			if (j <= 7)
			{
				flash_info[i].start[j] = CFG_FLASH_BASE +
				                         i * FLASH_BANK_SIZE +
				                         j * PARAM_SECT_SIZE;
			}
			else
			{
				flash_info[i].start[j] = CFG_FLASH_BASE +
				                         i * FLASH_BANK_SIZE +
				                         (j - 7)*MAIN_SECT_SIZE;
			}
		}
		size += flash_info[i].size;
	}

	/* Protect monitor and environment sectors
	 */
#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
#if CFG_MONITOR_BASE >= CFG_FLASH_BASE + FLASH_BANK_SIZE
	flash_protect(FLAG_PROTECT_SET,
	              CFG_MONITOR_BASE,
	              CFG_MONITOR_BASE + CFG_MONITOR_LEN - 1,
	              &flash_info[1]);
#else
	flash_protect(FLAG_PROTECT_SET,
	              CFG_MONITOR_BASE,
	              CFG_MONITOR_BASE + CFG_MONITOR_LEN - 1,
	              &flash_info[0]);
#endif
#endif

#if (CFG_ENV_IS_IN_FLASH == 1) && defined(CFG_ENV_ADDR)
#if CFG_ENV_ADDR >= CFG_FLASH_BASE + FLASH_BANK_SIZE
	flash_protect(FLAG_PROTECT_SET,
	              CFG_ENV_ADDR,
	              CFG_ENV_ADDR + CFG_ENV_SIZE - 1,
	              &flash_info[1]);
#else
	flash_protect(FLAG_PROTECT_SET,
	              CFG_ENV_ADDR,
	              CFG_ENV_ADDR + CFG_ENV_SIZE - 1,
	              &flash_info[0]);
#endif
#endif

	Done:
	return size;
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
	int i;

	switch (info->flash_id & FLASH_VENDMASK)
	{
		case (INTEL_MANUFACT & FLASH_VENDMASK):
			printf("Intel: ");
		break;
		default:
			printf("Unknown Vendor ");
			break;
	}

	switch (info->flash_id & FLASH_TYPEMASK)
	{
		case 0x88F4:
			printf("28F160F3B (16Mbit)\n");
			break;
		default:
			printf("Unknown Chip Type\n");
			goto Done;
			break;
	}

	printf("  Size: %ld MB in %d Sectors\n",
	       info->size >> 20, info->sector_count);

	printf("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; i++)
	{
		if ((i % 5) == 0)
		{
			printf ("\n   ");
		}
		printf (" %08lX%s", info->start[i],
		        info->protect[i] ? " (RO)" : "     ");
	}
	printf ("\n");

	Done:
}

/*-----------------------------------------------------------------------
 */

void	flash_erase (flash_info_t *info, int s_first, int s_last)
{
	int flag, prot, sect;
	ulong start, now, last;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return;
	}

	if ((info->flash_id & FLASH_VENDMASK) !=
	    (INTEL_MANUFACT & FLASH_VENDMASK)) {
		printf ("Can erase only Intel flash types - aborted\n");
		return;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	start = get_timer (0);
	last  = start;
	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			vu_long *addr = (vu_long *)(info->start[sect]);
			unsigned long status;

			/* Disable interrupts which might cause a timeout
			 * here.
			 */
			flag = disable_interrupts();

			*addr = 0x00500050;	/* clear status register */
			*addr = 0x00200020;	/* erase setup */
			*addr = 0x00D000D0;	/* erase confirm */

			/* re-enable interrupts if necessary */
			if (flag)
				enable_interrupts();

			/* wait at least 80us - let's wait 1 ms */
			udelay (1000);

			while (((status = *addr) & 0x00800080) != 0x00800080) {
				if ((now=get_timer(start)) >
				           CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
					*addr = 0x00B000B0; /* suspend erase */
					*addr = 0x00FF00FF; /* to read mode  */
					return;
				}

				/* show that we're waiting */
				if ((now - last) > 1000) {  /* every second  */
					putc ('.');
					last = now;
				}
			}

			*addr = 0x00FF00FF;	/* reset to read mode */
		}
	}
	printf (" done\n");
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 4 - Flash not identified
 */

#define	FLASH_WIDTH	8	/* flash bus width in bytes */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong wp, msr;
	int l, rc;
	unsigned char data[8];
	u_long *datah = (u_long *)&data[0];
	u_long *datal = (u_long *)&data[4];

	if (info->flash_id == FLASH_UNKNOWN) {
		return 4;
	}

	wp = (addr & ~(FLASH_WIDTH-1));	/* get lower aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
#if 0
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<FLASH_WIDTH && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<FLASH_WIDTH; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}

		if ((rc = write_data(info, wp, data)) != 0) {
			return (rc);
		}
		wp += FLASH_WIDTH;
#else
		return ERR_ALIGN;
#endif
	}

#if 1
	if (cnt % FLASH_WIDTH != 0)
	{
		return ERR_ALIGN;
	}
#endif

	/*
	 * handle FLASH_WIDTH aligned part
	 */
	msr = get_msr();
	set_msr(msr | MSR_FP);
	while (cnt >= FLASH_WIDTH) {
		*datah = *(u_long *)src;
		*datal = *(u_long *)(src + 4);
		if ((rc = write_data(info, wp, data)) != 0) {
			set_msr(msr);
			return (rc);
		}
		wp  += FLASH_WIDTH;
		cnt -= FLASH_WIDTH;
		src += FLASH_WIDTH;
	}
	set_msr(msr);

	if (cnt == 0) {
		return (0);
	}

#if 0
	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<FLASH_WIDTH && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i<FLASH_WIDTH; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}

	return (write_data(info, wp, data));
#else
	return 0;
#endif
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_data (flash_info_t *info, ulong dest, uchar *data)
{
	vu_long *addr = (vu_long *)dest;
	ulong status;
	ulong start;
	ulong *hdata = (ulong *)&data[0];
	ulong *ldata = (ulong *)&data[4];
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*addr & *hdata) != *hdata ||
	    (*(addr + 1) & *ldata) != *ldata) {
		return (2);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	*addr = 0x00400040;		/* write setup */
	write_via_fpu(addr, data);

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	start = get_timer (0);

	while ((((status = *addr) & 0x00800080) != 0x00800080) ||
	      ((*(addr + 1)) & 0x00800080) != 0x00800080) {
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			*addr = 0x00FF00FF;	/* restore read mode */
			return (1);
		}
	}

	*addr = 0x00FF00FF;	/* restore read mode */

	return (0);
}

/*-----------------------------------------------------------------------
 */
static void write_via_fpu(volatile ulong *addr, uchar *data)
{
__asm__ __volatile__ (
	"lfd  1, 0(%0)\n
	 stfd 1, 0(%1)\n"
	 :
	 : "r" (data), "r" (addr)
);
}
/*-----------------------------------------------------------------------
 */
static __inline__ unsigned long get_msr(void)
{
    unsigned long msr;

    asm volatile("mfmsr %0" : "=r" (msr) :);
    return msr;
}

static __inline__ void set_msr(unsigned long msr)
{
    asm volatile("mtmsr %0" : : "r" (msr));
}
